/*
 * The contents of this file are subject to the terms
 * of the Common Development and Distribution License
 * (the "License").  You may not use this file except
 * in compliance with the License.
 *
 * You can obtain a copy of the license at
 * https://jwsdp.dev.java.net/CDDLv1.0.html
 * See the License for the specific language governing
 * permissions and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL
 * HEADER in each file and include the License file at
 * https://jwsdp.dev.java.net/CDDLv1.0.html  If applicable,
 * add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your
 * own identifying information: Portions Copyright [yyyy]
 * [name of copyright owner]
 */
/*
 * $Id: AssertionIDReference.java,v 1.5 2007/01/08 16:05:56 shyam_rao Exp $
 */

/*
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package com.sun.xml.wss.saml.assertion.saml11.jaxb10;

import com.sun.xml.wss.saml.SAMLException;

import com.sun.xml.wss.logging.LogDomainConstants;
import java.util.Random;
import org.w3c.dom.Element;
import java.util.logging.Logger;

import javax.xml.bind.JAXBContext;

import com.sun.xml.wss.saml.internal.saml11.jaxb10.impl.AssertionIDReferenceImpl;
import com.sun.xml.wss.saml.util.SAMLJAXBUtil;

/**
 * <code>AssertionIDReference</code> element makes reference to a SAML
 * assertion.
 */
public class AssertionIDReference extends com.sun.xml.wss.saml.internal.saml11.jaxb10.impl.AssertionIDReferenceImpl 
    implements com.sun.xml.wss.saml.AssertionIDReference {
    
    protected static final Logger log = Logger.getLogger(
            LogDomainConstants.WSS_API_DOMAIN,
            LogDomainConstants.WSS_API_DOMAIN_BUNDLE);

    /**
     *Default constructor of <code>AssertionIDReferenceBase</code>
     */
    public AssertionIDReference() {
        super();
        Random rnd = new Random();
        int intRandom = rnd.nextInt();
        _Value = "XWSSSAML-"+String.valueOf(System.currentTimeMillis())+String.valueOf(intRandom);
    }

    /**
     * Constructs an <code>AssertionIDReference</code> element from an existing
     * XML block.
     *
     * @param element representing a DOM tree element.
     * @exception SAMLException if there is an error in the sender or in the
     *            element definition.
     */
    public static AssertionIDReferenceImpl fromElement(Element element) throws SAMLException {
        try {
            JAXBContext jc =
                    SAMLJAXBUtil.getJAXBContext();
            javax.xml.bind.Unmarshaller u = jc.createUnmarshaller();
            return (AssertionIDReferenceImpl)u.unmarshal(element);
        } catch ( Exception ex) {
            throw new SAMLException(ex.getMessage());
        }
    }

    /**
     *This is a constructor of <code>AssertionIDReference</code>.
     *@param id A String representing an existing assertion id.
     */
    public AssertionIDReference(String id) {
        super(id);
    }    
}
